from flask import render_template
from realnet.resource.items.items import Items

class Forms(Items):
    
    def get_query(self, module, account, query, parent_item=None):
        return {'types':['Form'], 'any_level': 'true'}


import json
from flask import jsonify, render_template, redirect
from realnet.resource.items.items import Items
from realnet.core.type import Item, Instance

class Types(Items):
    
    def match(self, instance, query):
        if query and instance:
            # todo proper check for inheritance
            types = set(query.get('types',[]))
            for type in types:
                if instance.type.is_derived_from(type) or type == 'Item':
                    return True
        else:
            return True

    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        if parent_item and parent_item.instance.type.name != 'Types':
            instances = [i for i in parent_item.instance.type.instances if self.match(i, query)]
            for i in instances:
                if i.attributes:
                    attrs = dict(i.attributes)
                    attrs['resource'] = 'types'
                    attrs['forms'] = [
                        { 'name': 'create', 'type': 'FormItem', 'attributes': { 'form': 'InstanceCreateForm', 'path': 'types' } },
                        { 'name' : 'edit', 'type': 'FormItem', 'attributes': { 'form': 'InstanceEditForm', 'path': 'types' }  },
                        { 'name': 'delete', 'type': 'FormItem', 'attributes': { 'form': 'DeleteForm', 'path': 'types' }  }
                    ]
                    i.attributes = attrs
                else:
                    i.attributes = {'resource': 'types'}
            return instances
        else:
            types = [t for t in module.get_types()]
            for t in types:
                attrs = dict(t.attributes)
                attrs['resource'] = 'types'
                t.attributes = attrs
            return types
        
    def get_template_args(self, module, endpoint, args, path):
        return super().get_template_args(module, endpoint, args, path)

    def get_item(self, module, endpoint, account, args, path):
        type = module.get_type_by_id(path)
        if not type:
            instance = module.get_instance_by_id(path)
            return Item(account.id, account.org.id, instance, instance.id, instance.name)
        else:
            return Item(account.id, account.org.id, Instance(type.id, type, type.name), type.id, type.name)

    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        if 'type' in args:
            if args['type'] == 'Attribute':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, path, attrs['parent_id'])
                    if item:
                        if module.can_account_write_item(account, item):
                            params = dict()
                            current_attrs = dict(item.attributes)
                            current_attrs[attrs['name']] = attrs['value']
                            params['attributes'] = current_attrs
                            module.update_type(attrs['parent_id'], **params)
                    if 'active_view' in attrs:
                        return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                    else:
                        return redirect('/types/{}'.format(attrs['parent_id']))
                    
            elif args['type'].endswith('View'):
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            view = dict()
                            if 'name' in args:
                                view['name'] = args['name']
                            if 'attributes' in args:
                                view['attributes'] = args.get('attributes',dict())
                            if 'type' in args:
                                view['type'] = args.get('type')
                            
                            views = item.attributes.get('views', []) + [view]
                            current_attrs = dict(item.attributes)
                            current_attrs['views'] = views
                            params = {'attributes': current_attrs }
                            module.update_type(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/types/{}'.format(attrs['parent_id']))
            elif args['type'] == 'FormItem':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            form = dict()
                            form['type'] = 'FormItem'
                            form['attributes'] = dict()
                            if 'name' in args:
                                form['name'] = args['name']
                            if 'path' in args:
                                form['attributes']['path'] = args['path']
                            if 'form' in args:
                                form['attributes']['form'] = args['form']
                            
                            forms = item.attributes.get('forms', []) + [form]
                            current_attrs = dict(item.attributes)
                            current_attrs['forms'] = forms
                            params = {'attributes': current_attrs }
                            module.update_type(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/types/{}'.format(attrs['parent_id']))
            elif args['type'] == 'MenuItem':
                attrs = dict(args)
                if 'parent_id' in attrs:
                    account = module.get_account()
                    item = self.get_item(module, endpoint, account, attrs, attrs['parent_id']);
                    if item:
                        if module.can_account_write_item(account, item):
                            form = dict()
                            form['type'] = 'MenuItem'
                            form['attributes'] = dict()
                            if 'name' in args:
                                form['name'] = args['name']
                            if 'path' in args:
                                form['attributes']['path'] = args['path']
                            if 'form' in args:
                                form['attributes']['form'] = args['form']
                            if 'icon' in args:
                                form['attributes']['icon'] = args['icon']
                            
                            forms = item.attributes.get('menu', []) + [form]
                            current_attrs = dict(item.attributes)
                            current_attrs['menu'] = forms
                            params = {'attributes': current_attrs }
                            module.update_type(attrs['parent_id'], **params)
                            if 'active_view' in attrs:
                                return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                            else:
                                return redirect('/types/{}'.format(attrs['parent_id']))
            elif 'parent_id' in args:
                attrs = dict(args)
                attrs['parent_type_id'] = attrs['parent_id']
                item = module.create_instance(**attrs)
                if 'active_view' in attrs:
                    return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                else:
                    return redirect('/types/{}'.format(attrs['parent_id']))
        else:
            attrs = dict()
            for k,v in args.items():
                if k not in set(['name', 'base']):
                    if not 'attributes' in attrs:
                        attrs['attributes'] = dict()
                    attrs['attributes'][k] = v
                else:
                    attrs[k] = v

            if not 'attributes' in attrs:
                attrs['attributes'] = dict()
            
            if not 'resource' in attrs['attributes'] or not attrs['attributes']['resource']:
                attrs['attributes']['resource'] = 'types'

            item = module.create_type(**attrs)
            
        if content_type == 'application/json':
            return jsonify(item.to_dict())
        else:
            return self.render_item(module, endpoint, args, path, content_type)

    def put(self, module, endpoint, args, path=None, content_type='text/html'):
        attrs = dict(args)
        id = None
        is_instance = False
        if path:
            id = path
        elif 'id' in attrs:
            id = attrs['id']
        elif 'parent_id' in attrs:
            id = attrs['parent_id']
            
        if 'parent_id' in attrs:
            is_instance = True

        if id:
            account = module.get_account()
            item = self.get_item(module, account, attrs, id);
            if item:
                if module.can_account_write_item(account, item):
                    if 'type' in args:
                        if args['type'].endswith('View'):
                            if 'name' in args:
                                name = args['name'].lower()
                                views = [] 
                                for view in item.attributes.get('views', []):
                                    if view['name'].lower() != name:
                                        views.append(view)
                                    else:
                                        views.append({'name': view['name'], 'type': args['type'], 'attributes': args.get('attributes',{})})
                                current_attrs = dict(item.attributes)
                                current_attrs['views'] = views
                                params = {'attributes': current_attrs }
                                module.update_type(id, **params)
                                if 'parent_id' in attrs:
                                    if 'active_view' in attrs:
                                        return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                                    else:
                                        return redirect('/types/{}'.format(attrs['parent_id']))

                                return redirect('/types/{}'.format(id))
                        elif args['type'] == 'FormItem':
                            if 'name' in args:
                                name = args['name'].lower()
                                forms = [] 
                                for form in item.attributes.get('forms', []):
                                    if form['name'].lower() != name:
                                        forms.append(form)
                                    else:
                                        forms.append({'name': form['name'], 'type': form['type'], 'attributes': {'form': args['form'], 'path': args['path']}})
                                current_attrs = dict(item.attributes)
                                current_attrs['forms'] = forms
                                params = {'attributes': current_attrs }
                                module.update_type(id, **params)
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(id, attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(id))
                        elif args['type'] == 'MenuItem':
                            if 'name' in args:
                                name = args['name'].lower()
                                forms = [] 
                                for form in item.attributes.get('menu', []):
                                    if form['name'].lower() != name:
                                        forms.append(form)
                                    else:
                                        forms.append({'name': form['name'], 'type': form['type'], 'attributes': {'form': args['form'], 'path': args['path'], 'icon': args['icon']}})
                                current_attrs = dict(item.attributes)
                                current_attrs['menu'] = forms
                                params = {'attributes': current_attrs }
                                module.update_type(id, **params)
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(id, attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(id))
                        elif args['type'] == 'Attribute':
                            if 'name' in args and 'value' in args:
                                attrs = dict(item.attributes)
                                try:
                                    attrs[args['name']] = json.loads(args['value'])
                                except Exception as e:
                                    attrs[args['name']] = args['value']
                                params = dict()
                                params['attributes'] = attrs
                                module.update_type(id, **params)
                                if 'active_view' in args:
                                    return redirect('/types/{}?view={}'.format(id, args['active_view']))
                                else:
                                    return redirect('/types/{}'.format(id))


                    if id:
                        params = dict()
                        if is_instance:
                            if 'name' in args:
                                params['name'] = args['name']
                            if 'type' in args:
                                params['type'] = args['type']

                            module.update_instance(id, **params)

                            if 'parent_id' in attrs:
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(attrs['parent_id']))
                        else:
                            current_attrs = dict(item.attributes)
                            
                            for k,v in args.items():
                                if k == 'name' or k == 'base':
                                    params[k] = args[k]
                                else:
                                    current_attrs[k] = args[k]
                            

                            params['attributes'] = current_attrs
                            module.update_type(id, **params)

                            if 'active_view' in attrs:
                                return redirect('/types/{}?view={}'.format(id, attrs['active_view']))
                            else:
                                return redirect('/types/{}'.format(id))
                                
                    return redirect('/types/{}'.format(id))
                    
        # module.delete_type(args['id'])
        if 'parent_id' in attrs:
            if 'active_view' in attrs:
                return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
            else:
                return redirect('/types/{}'.format(attrs['parent_id']))
            
        return redirect('/types/{}'.format(id))

    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        attrs = dict(args)
        id = None
        if path:
            id = path
        elif 'id' in attrs:
            id = attrs['id']
        elif 'parent_id' in attrs:
            id = attrs['parent_id']

        if id:
            account = module.get_account()
            item = self.get_item(module, endpoint, account, attrs, id);
            if item:
                if module.can_account_write_item(account, item):
                    if 'type' in args:
                        if args['type'].endswith('View'):
                            if 'name' in args:
                                name = args['name'].lower()
                                views = [] 
                                for view in item.attributes.get('views', []):
                                    if view['name'].lower() != name:
                                        views.append(view)
                                current_attrs = dict(item.attributes)
                                current_attrs['views'] = views
                                params = {'attributes': current_attrs }
                                module.update_type(attrs['id'], **params)
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(attrs['id'], attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(attrs['id']))
                        if args['type'] == 'FormItem':
                            if 'name' in args:
                                name = args['name'].lower()
                                forms = [] 
                                for form in item.attributes.get('forms', []):
                                    if form['name'].lower() != name:
                                        forms.append(form)
                                current_attrs = dict(item.attributes)
                                current_attrs['forms'] = forms
                                params = {'attributes': current_attrs }
                                module.update_type(id, **params)
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(id, attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(id))
                        elif args['type'] == 'MenuItem':
                            if 'name' in args:
                                name = args['name'].lower()
                                forms = [] 
                                for form in item.attributes.get('menu', []):
                                    if form['name'].lower() != name:
                                        forms.append(form)
                                current_attrs = dict(item.attributes)
                                current_attrs['menu'] = forms
                                params = {'attributes': current_attrs }
                                module.update_type(id, **params)
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(id, attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(id))
                        elif args['type'] == 'Attribute':
                            if 'name' in args:
                                attrs = dict(item.attributes)
                                if args['name'] in attrs:
                                    del attrs[args['name']]
                                    params = dict()
                                    params['attributes'] = attrs
                                    module.update_type(id, **params)
                                    if 'active_view' in args:
                                        return redirect('/types/{}?view={}'.format(id, args['active_view']))
                                    else:
                                        return redirect('/types/{}'.format(id))
                        else:
                            if id:
                                if attrs['id'] == attrs['parent_id']:
                                    module.delete_type(id)
                                    return redirect('/types')
                                else:
                                    module.delete_instance(id)
                            if 'parent_id' in attrs:
                                if 'active_view' in attrs:
                                    return redirect('/types/{}?view={}'.format(attrs['parent_id'], attrs['active_view']))
                                else:
                                    return redirect('/types/{}'.format(attrs['parent_id']))
                                    
        # module.delete_type(args['id'])
        return redirect('/types/{}'.format(id))


import importlib.util
import sys

from realnet.core.provider import ResourceProvider
from realnet.core.type import Func


class GenericResourceProvider(ResourceProvider):

    def __init__(self):
        pass

    def get_resources(self, module):
        # account = module.get_account()
        return [] #[Endpoint(e) for e in module.find_items({'types': ['Endpoint'], 'children': 'true'}) if self.acl.can_account_read_item(account, e)]
    
    def get_resource(self, module, name):
        resources = [r for r in module.find_items({ 'types': ['Resource'], 
                                                    'name': name, 
                                                    'any_level': 'true'})]
        return next(iter(resources), None)

    def get_resource_method(self, module, endpoint, resource_name, method_name):
        account = module.get_account()
        resources = [r for r in module.find_items({ 'types': ['Resource'], 
                                                    'name': resource_name, 
                                                    'any_level': 'true',
                                                    'keys': ['module'], 
                                                    'values': ['true']})]
        resource_item = next(iter(resources), None)
        if resource_item:
            module_path = resource_item.attributes.get('module_path')
            module_class_name = resource_item.attributes.get('module_class')

            code = resource_item.attributes.get('code')
            if not code:
                resource_data = None
                if module_path:
                    resource_module = importlib.import_module(module_path)
                    resource_class = getattr(resource_module, module_class_name)
                    resource_instance = resource_class()
                    func = getattr(resource_instance, method_name, None)
                    if callable(func):
                        return Func(callback=lambda module, endpoint, args, path, content_type : func(module, endpoint, args, path, content_type))
                else:
                    resource_data = module.get_data(resource_item.id)
                    if resource_data:
                        code = resource_data.bytes.decode('utf-8')
            
            if code:
                d = dict(globals())
                l = dict(locals())
                res = exec(code, d, l)
                cls = l.get('Resource')
                if cls:
                    resource_instance = cls()
                    func = getattr(resource_instance, method_name, None)
                    if callable(func):
                        return Func(callback=lambda module, endpoint, args, path, content_type : func(module, endpoint, args, path, content_type))
        return None


import uuid

from flask import render_template, redirect, current_app
from realnet.resource.items.items import Items
from realnet.core.type import Instance, Item

class Clients(Items):
    
    def item_from_client(self, client, client_type):
        instance = Instance(client.id, client_type, client.name, {})
        return Item(client.id, client.org.id, instance, client.id, client.name, dict(), [])
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        tbn = {t.name:t for t in module.get_types()}
        account = module.get_account()
        contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
        clients = contextProvider.get_org_clients(account.org.id)
        return [self.item_from_client(r, tbn['Client']) for r in clients]
    
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'account_id' in args:
                pass
            else:
                request = dict()
                
                request['name'] = args.get('name')
                request['uri'] = args.get('uri')
                request['scope'] = args.get('scope')
                request['auth_method'] = args.get('auth_method')
                request['grant_types'] = args.get('grant_types')
                request['redirect_uris'] = args.get('redirect_uris')
                request['response_types'] = args.get('response_types')
                request['org_id'] = account.org.id
                
                auth = module.create_org_client(**request)
                return redirect('/clients')
            
        return self.render_item(module, endpoint, args, path, content_type)
    
    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'id' in args:
                pass
                # module.remove_role_app(args['parent_id'], args['id'])
                # return redirect('/roles/{}'.format(args['parent_id']))
            elif 'id' in args:
                module.remove_org_client(account.org.id, args['id'])
            elif 'name' in args:
                module.remove_org_client(account.org.id, args['name'])

        return self.render_item(module, endpoint, args, path, content_type)


import uuid

from flask import render_template, redirect, current_app
from realnet.resource.items.items import Items
from realnet.core.type import Instance, Item

class Auths(Items):

    def item_from_auth(self, auth, auth_type):
        instance = Instance(auth.id, auth_type, auth.name, {"url":auth.url})
        return Item(auth.id, auth.org_id, instance, auth.id, auth.name, dict(), [])
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        tbn = {t.name:t for t in module.get_types()}
        account = module.get_account()
        contextProvider = current_app.config['REALNET_CONTEXT_PROVIDER']
        auths = contextProvider.get_org_authenticators(account.org.id)
        return [self.item_from_auth(r, tbn['Auth']) for r in auths]
    
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'account_id' in args:
                pass
            else:
                request = dict()
                
                request['name'] = args.get('name')
                request['api_base_url'] = args.get('api_base_url')
                request['request_token_url'] = args.get('request_token_url')
                request['access_token_url'] = args.get('access_token_url')
                request['authorize_url'] = args.get('authorize_url')
                request['client_kwargs'] = args.get('client_kwargs')
                request['client_id'] = args.get('client_id')
                request['client_secret'] = args.get('client_secret')
                request['userinfo_endpoint'] = args.get('userinfo_endpoint')
                request['server_metadata_url'] = args.get('server_metadata_url')
                request['redirect_url'] = args.get('redirect_url')
                request['scope'] = args.get('scope')
                request['org_id'] = account.org.id
                
                auth = module.create_org_auth(**request)
                return redirect('/auths')
            
        return self.render_item(module, endpoint, args, path, content_type)
    
    def delete(self, module, endpoint, args, path=None, content_type='text/html'):
        account = module.get_account()
        if account.is_superuser() or account.is_admin():
            if 'parent_id' in args and 'id' in args:
                pass
                # module.remove_role_app(args['parent_id'], args['id'])
                # return redirect('/roles/{}'.format(args['parent_id']))
            elif 'id' in args:
                module.remove_org_auth(account.org.id, args['id'])
            elif 'name' in args:
                module.remove_org_auth(account.org.id, args['name'])

        return self.render_item(module, endpoint, args, path, content_type)


import os
import sys
from flask import render_template, jsonify
from realnet.resource.items.items import Items

class Apps(Items):
    
    def get_query(self, module, account, query, parent_item=None):
        return {'types':['App']}
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        return module.get_apps(module)
    
    def post(self, module, endpoint, args, path=None, content_type='text/html'):
        if args:
            for key in args:
                if key != 'types':
                    val = args[key]
                    if isinstance(val, list):
                        args[key] = val[0]
        # app_type = module.create_type(name=args.get('name'), base='App')
        
        resource = module.get_resource(module, args.get('resource'))
        if not resource:
            resource_item = module.create_item(name=args.get('resource'), type='Resource', attributes={'module':'true'}, public='true')
            with open(os.path.join(os.path.abspath(os.path.join(os.path.dirname(sys.modules[__name__].__file__), os.pardir)), '../static/initialization/resource.py'), 'r') as f:
                module.update_data(resource_item.id, f.read())

        app_endpoint = module.get_endpoint(module, args.get('endpoint'))
        if not app_endpoint:
            endpoint_item = module.create_item(name=args.get('endpoint'), type='Endpoint', attributes={'path':args.get('endpoint'), 'resource':args.get('resource')}, public='true')
        
        app_item = module.create_item(name=args.get('name'), type='App', attributes={'icon':args.get('icon')})
                
        if content_type == 'application/json':
            return jsonify(app_item.to_dict())
        else:
            return self.render_item(module, endpoint, args, path, content_type)

    def get_template1(self, endpoint, module, args, path):
        if not path:
            return "apps.html"
        else:
            parts = path.split('/')
            tail = parts[-1]
            if tail == 'attributes':
                if len(parts) == 2:
                    return "app_attributes.html"
                else:
                    return "app_view_attributes.html"
            else:
                return "app_views.html"

    def get_template_args1(self, module, args, path):
        if not path:
            return super().get_template_args(module, args, path)
        else:
            parts = path.split('/')
            tail = parts[-1]
            if tail == 'attributes':
                if len(parts) == 2:
                    account = module.get_account()
                    org = module.get_org()
                    apps = module.get_apps(module)
                    app = next((a for a in apps if a.name.lower() == 'apps'), None)
                    return {'account': account,
                            'org': org,
                            'apps': apps,
                            'app': app,
                            'attributes': [{"name": a[0], "value": a[1]} for a in app.attributes.items()]}
                else:
                    account = module.get_account()
                    org = module.get_org()
                    apps = module.get_apps(module)
                    app = next((a for a in apps if a.name.lower() == 'apps'), None)
                    return {'account': account,
                            'org': org,
                            'apps': apps,
                            'app': app,
                            'attributes': [{"name": a[0], "value": a[1]} for a in app.attributes.items()]}
            else:
                account = module.get_account()
                org = module.get_org()
                apps = module.get_apps(module)
                app = next((a for a in apps if a.name.lower() == 'apps'), None)
                views = app.attributes.get('views',[])
                tbn = {t.name:t for t in module.get_types()}
                types_by_id = {t.id:t for t in tbn.values()}
                type_ids = set([t.id for t in tbn.values() if t.name == 'View'])
                types = [types_by_id[t] for t in module.get_derived_types(type_ids)]
                return {'account': account,
                        'org': org,
                        'apps': apps,
                        'app': app,
                        'views': views,
                        'types': types}


from flask import render_template
from realnet.resource.gltf.export import build_gltf_from_items
from realnet.resource.items.items import Items

class Gltf(Items):
    
    def render_item(self, module, endpoint, args, path=None, content_type='text/html'):
        if content_type == 'application/json' or content_type == '*/*':
            account = module.get_account()
            query = self.get_query(module, args | {'types': ['Item']}, path)
            items = self.get_items(module, endpoint, args, path, account, query)
            return build_gltf_from_items(module, items).to_json()
        else:
            return self.render_items_html(module, endpoint, args, path)


import uuid

from sqlalchemy.sql import func, or_

from werkzeug.security import gen_salt

from realnet.core.provider import OrgProvider, GroupProvider, AclProvider,AccountProvider, AppProvider
from realnet.core.type import Account, Authenticator, Org, Group, Client
from realnet.core.acl import AclType
from .utility import get_types_by_name, item_model_to_item, get_derived_types
from .models import Account as AccountModel, Org as OrgModel, Item as ItemModel, Group as GroupModel, GroupRoleType, AccountGroup as AccountGroupModel, Authenticator as AuthenticatorModel, Client as ClientModel, session as db, create_client

class SqlOrgProvider(OrgProvider, GroupProvider, AclProvider, AccountProvider, AppProvider):
    
    def __init__(self, org_id, account_id):
        self.org_id = org_id
        self.account_id = account_id

    def get_account_groups(self, account_id):
        account = db.query(AccountModel).filter(AccountModel.org_id == self.org_id, AccountModel.id == account_id).first()
        if account:
            return [Group(account_group.group.id, 
                        account_group.group.name, 
                        Org(account_group.group.org.id, account_group.group.org.name)) for account_group in account.groups]
        else:
            return []
    
    def add_account_group(self, account_id, group_id):
        account = db.query(AccountModel).filter(AccountModel.org_id == self.org_id, AccountModel.id == account_id).first()
        group = db.query(GroupModel).filter(GroupModel.org_id == self.org_id, GroupModel.id == group_id).first()
        if account and group:
            account_group = db.query(AccountGroupModel).filter(AccountGroupModel.org_id == self.org_id, AccountGroupModel.account_id == account_id, AccountGroupModel.group_id == group_id).first()
            if not account_group:
                account_group = AccountGroupModel(id=str(uuid.uuid4()),account_id=account_id, group_id=group_id, org_id=self.org_id, role_type=GroupRoleType.member)
                db.add(account_group)
                db.commit()
                
            return account_group.id

        return None
    
    def remove_account_group(self, account_id, group_id):
        account = db.query(AccountModel).filter(AccountModel.org_id == self.org_id, AccountModel.id == account_id).first()
        group = db.query(GroupModel).filter(GroupModel.org_id == self.org_id, GroupModel.id == group_id).first()
        if account and group:
            account_group = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account_id, AccountGroupModel.group_id == group_id).first()
            if account_group:
                db.delete(account_group)
                db.commit()

    def get_org_groups(self, org_id):
        return [Group(acc.id, acc.name, Org(acc.org.id, acc.org.name)) for acc in db.query(GroupModel).filter(GroupModel.org_id == self.org_id).all()]
    
    def create_org_group(self, org_id, name):
        group = GroupModel(  id=str( uuid.uuid4()), 
                               name=name, 
                               org_id=org_id)
        db.add(group)
        db.commit()
        return Group(group.id, group.name, Org(group.org.id, group.org.name))

    def remove_org_group(self, org_id, name):
        group = db.query(GroupModel).filter(GroupModel.org_id == org_id, or_(GroupModel.name == name, GroupModel.id == name)).first()
        if group:
            db.delete(group)
            db.commit()

    def create_org_auth(self, 
                        org_id, 
                        name,
                        api_base_url,
                        request_token_url,
                        access_token_url,
                        authorize_url,
                        client_kwargs,
                        client_id,
                        client_secret,
                        userinfo_endpoint,
                        server_metadata_url,
                        redirect_url,
                        scope):
        auth = AuthenticatorModel(  id=str( uuid.uuid4()),
                                    name=name,
                                    api_base_url=api_base_url,
                                    request_token_url=request_token_url,
                                    access_token_url=access_token_url,
                                    authorize_url=authorize_url,
                                    client_kwargs=client_kwargs,
                                    client_id=client_id,
                                    client_secret=client_secret,
                                    userinfo_endpoint=userinfo_endpoint,
                                    server_metadata_url=server_metadata_url,
                                    redirect_url=redirect_url,
                                    scope=scope,
                                    org_id=org_id)
        db.add(auth)
        db.commit()

        return Authenticator(auth.id, auth.name, auth.org_id, auth.api_base_url)

    def remove_org_auth(self, org_id, name):
        auth = db.query(AuthenticatorModel).filter(AuthenticatorModel.org_id == org_id, or_(AuthenticatorModel.name == name, AuthenticatorModel.id == name)).first()
        if auth:
            db.delete(auth)
            db.commit()

    def create_org_client(self, org_id, name, uri, grant_types, redirect_uris, response_types, scope, auth_method):
        client_id = gen_salt(24)
        client = create_client(name=name,
                            client_id=client_id,
                            uri=uri,
                            grant_types=grant_types,
                            redirect_uris=redirect_uris,
                            response_types=response_types,
                            scope=scope,
                            auth_method=auth_method,
                            org_id=org_id)
        return Client(client_id, 
                client.name, 
                client.org, 
                {   
                    'uri': client.client_uri, 
                    'grant_types': client.grant_types, 
                    'redirect_uris': client.redirect_uris, 
                    'response_types': client.response_types, 
                    'scope': client.scope, 
                    'auth_method': auth_method
                }
            )
    

    def remove_org_client(self, org_id, name):
        client = db.query(ClientModel).filter(ClientModel.org_id == org_id, or_(ClientModel.name == name, ClientModel.id == name)).first()
        if client:
            db.delete(client)
            db.commit()
    
    def get_org_accounts(self, org_id):
        return [Account(acc.id, acc.username, Org(acc.org.id, acc.org.name), acc.org_role_type) for acc in db.query(AccountModel).filter(AccountModel.org_id == self.org_id).all()]

    # acl provider

    def is_item_public(self, item):
    
        if [acl for acl in item.acls if acl.type == AclType.public]:
            return True

        return False

    def can_account_execute_item(self, account, item):
        
        if [acl for acl in item.acls if acl.type.value == AclType.public.value]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.user.value and acl.target_id == account.id and acl.permission and  'e' in acl.permission]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if acl.type.value == AclType.org.value and acl.org_id == account.org.id and  acl.permission and 'e' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.group.value and acl.target_id in group_ids and  acl.permission and 'e' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_message_item(self, account, item):

        if [acl for acl in item.acls if acl.type.value == AclType.public.value]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.user.value and acl.target_id == account.id and  acl.permission and 'm' in acl.permission]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if acl.type.value == AclType.org.value and acl.org_id == account.org.id and acl.permission and  'm' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.group.value and acl.target_id in group_ids and acl.permission and  'm' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_read_item(self, account, item):

        if [acl for acl in item.acls if acl.type.value == AclType.public.value]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.user.value and acl.target_id == account.id and acl.permission and ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if acl.type.value == AclType.group.value and acl.target_id in group_ids and acl.permission and  ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.org.value and acl.org_id == account.org.id and acl.permission and  ('r' in acl.permission or 'w' in acl.permission)]:
            return True

        if item.owner_id == account.id:
            return True


    def can_account_write_item(self, account, item):

        if [acl for acl in item.acls if
            acl.type == AclType.user and acl.target_id == account.id and acl.permission and 'w' in acl.permission]:
            return True

        account_groups = db.query(AccountGroupModel).filter(AccountGroupModel.account_id == account.id).all()
        group_ids = []
        for account_group in account_groups:
            group_ids.append(account_group.group_id)

        if [acl for acl in item.acls if
            acl.type.value == AclType.group.value and acl.target_id in group_ids and acl.permission and 'w' in acl.permission]:
            return True

        if [acl for acl in item.acls if acl.type.value == AclType.org.value and acl.org_id == account.org.id and acl.permission and  'w' in acl.permission]:
            return True

        if item.owner_id == account.id:
            return True

    def can_account_delete_item(self, account, item):
        return item.owner_id == account.id and account.org.id == item.org_id

    # group provider

    def get_groups(self):
        pass

    # account provider

    def get_account(self):
        account = db.query(AccountModel).filter(AccountModel.id == self.account_id).first()
        if account:
            return Account(account.id, account.username, Org(account.org.id, account.org.name), account.org_role_type)
        return None

    def get_org(self):
        org = db.query(OrgModel).filter(OrgModel.id == self.org_id).first()
        if org:
            return Org(org.id, org.name)
        return None

    # app provider

    def get_apps(self, module):
        account = db.query(AccountModel).filter(AccountModel.id == self.account_id).first()
        if account:
            tbn = get_types_by_name(self.org_id)
            application_type = tbn.get('App')
            application_type_ids = set(get_derived_types(self.org_id, [application_type.id]) + [application_type.id])
            role_apps = [app.app for ar in account.roles for app in ar.role.apps ]
            owned_apps = db.query(ItemModel).filter(ItemModel.owner_id == self.account_id, ItemModel.type_id.in_(application_type_ids) ).all()
            app_ids = {app.id:app for app in role_apps}
            for owned_app in owned_apps:
                if not owned_app.id in app_ids:
                    role_apps.append(owned_app)
                    app_ids[owned_app.id] = owned_app

            return [item_model_to_item(self.org_id, app, tbn, False) for app in role_apps ]
        return []






import os
import random
import string
import requests
from dotenv import *
from flask import Flask, request, jsonify

from realnet.resource.items.items import Items
from realnet.core.type import Item, Instance

path = os.path.join(os.getcwd(), ".env")
if os.path.exists(path):
    load_dotenv(dotenv_path=path)

import re

class Chatgpt(Items):

    def call_chatgpt_api(self, prompt):
        CHATGPT_API_KEY = os.getenv('REALNET_CHATGPT_API_KEY', 'your_api_key_here')
        CHATGPT_API_URL = os.getenv('REALNET_CHATGPT_API_URL', 'https://api.openai.com/v1/engines/davinci-codex/completions')
        headers = {
            "Authorization": f"Bearer {CHATGPT_API_KEY}",
            "Content-Type": "application/json",
        }
        data = {
            "model": "gpt-3.5-turbo",
            "messages": [
                {
                    "role": "user",
                    "content": prompt,
                }
            ] # You can adjust this value as needed
        }
        response = requests.post(CHATGPT_API_URL, headers=headers, json=data)
        response.raise_for_status()
        return response.json()["choices"][0]["message"]["content"]

    def get_taxonomic_info(self, word):
        taxonomy_prompt = f"give me a ### delimited 4-tuple response where tuple elements correspond to the following specs in this order: t1. path last end of th corresponding Wikipedia url of the term {word} t2. the Wikipedia link corresponding to the term {word}, t3. list 50 most important attributes described in the wikipedia page text corresponding to term {word} in ('name': 'value') pairs embedded in [], t4. list of any relevant internet links for term {word} based on its Wikipedia page embedded as []. Write the tuple in the format t1###t2###t3###t4)"
        taxonomy_and_link_and_attributes_and_links = self.call_chatgpt_api(taxonomy_prompt)
        answers = taxonomy_and_link_and_attributes_and_links.strip().split('###')
        attributes = '[]'
        links = '[]'
        if len(answers) != 4:
            print(answers)
            taxonomy, wikimedia_link, attributes, links = '', '', '[]', '[]'
        else:
            taxonomy, wikimedia_link, attributes, links = answers

        # Process attributes and links
        attribute_list = []

        for string in attributes.strip('[]').split(','):  # remove the square brackets and split the string by comma
            string = string.strip("' (){}")  # remove any leading/trailing single quotes and whitespace
            if ':' in string:
                name, value = string.split(':', 1)
                attribute_list.append({"name": name.strip("' "), "value": value.strip("' ")})
        
        link_list = []
        for string in links.strip('[]').split(','):
            link = string.strip("' ")  # remove any leading/trailing single quotes and whitespace
            link_list.append(link)

        return taxonomy, wikimedia_link, attribute_list, link_list

    def starts_with_number_dot(self, word):
        pattern = r'^\d+\.' # regular expression pattern to match number followed by dot at the start of the word
        return bool(re.match(pattern, word))
    
    def in_whitelist(self, word):
        return not self.starts_with_number_dot(word) and all(char.isalnum() or char == '.' for char in word)

    def parse_response(self, module, response_text):
        words = response_text.split()
        objects = []
        for word in words:
            if self.in_whitelist(word):
                taxonomy, wikimedia_link, attributes, links = self.get_taxonomic_info(word)

                # Add wikimedia_link to attributes
                attributes.append({"name": "wikimedia_link", "value": wikimedia_link})
                attributes.append({"name": "links", "value": links})

                param_attributes = dict()
                for attribute in attributes:
                    param_attributes[attribute['name']] = attribute['value']

                param_attributes['resource'] = 'chatgpt'
                param_attributes['icon'] = 'square'

                physical_object = {
                    "name": word,
                    "type": taxonomy,
                    "attributes": param_attributes,
                }
                objects.append(self.get_or_create_item(module, physical_object))
        return objects
    
    def get_or_create_item(self, module, item):
        # Check if item already exists
        existing_item = module.find_items({'name': item['name']})
        if existing_item:
            return existing_item[0]
        else:
            # Create new item
            item_type = module.get_type_by_name(item['type'])
            if not item_type:
                item_type = module.create_type(name=item['type'], base='Type', attributes=item['attributes'])

            random_string = '' #''.join(random.choice(string.ascii_letters + string.digits) for _ in range(5))
            item_instance = module.create_instance(name=item['name'] + random_string, type=item_type.name)
            item = module.create_item(name=item['name'] + random_string, type=item_type.name, instance_id=item_instance.id)
            return item
    
    def get_items(self, module, endpoint, args, path, account, query, parent_item=None):
        input_message = args.get("name",[])
        if not input_message:
            return []

        chatgpt_response = self.call_chatgpt_api(input_message[0])
        items = self.parse_response(module, chatgpt_response)
        return items
        
        